#include "AST.h"
#include "ArrayAccessSet.h"
#include "code-util.h"
#include "optimize.h"

/* Given an array arr, return the source for its VarDecl. */
static TreeNode *arrayVarDecl(TreeNode *arr)
{
  Decl *d = arr->decl();
  if (!d->hasSource()) return NULL;
  return d->source();
}

MIVEset * & ArrayAccessSet::MIVEs(TreeNode *arr)
{
  return m[arrayVarDecl(arr)];
}

string ArrayAccessSet::OSRTempname(TreeNode *WRTloop, TreeNode *arr,
				   const MIVE *m)
{
  TreeNode *d = arrayVarDecl(arr);
  int i = find_or_append_MIVE(m, q[d]);
  return string("OSR_") + int2alpha(unique2((void *) WRTloop,
					    (void *) arr->decl()->source())) +
    "_" + int2alpha(i / 2);
}

const MIVE *ArrayAccessSet::baseMIVEForOSR(TreeNode *arr, const MIVE *m)
{
  TreeNode *d = arrayVarDecl(arr);
  if (d == NULL)
    return NULL;
  llist<const MIVE *> *l = q[d];
  if (l == NULL)
    q.erase(d);
  while (l != NULL)
    if (l->tail()->front()->equals(m))
      return l->front();
    else
      l = l->tail()->tail();
  return NULL;
}

/* x is the base, and we're going to do y off it. */
void ArrayAccessSet::adjoinOSR(TreeNode *t, TreeNode *WRTloop,
			       const MIVE *x, const MIVE *y)
{
  llist<const MIVE *> * & l = q[arrayVarDecl(t->array())];
  l = cons(x, cons(y, l));
}

void ArrayAccessSet::adjoinConstant(TreeNode *t, TreeNode *WRTloop)
{
  TreeNode *d = arrayVarDecl(t->array());
  TreeNode *indexdecl = t->index()->decl()->source();
  if (d != NULL && indexdecl != NULL) {
    treeSet *& s = c[d];
    if (s == NULL)
      s = new treeSet;
    ::adjoin(s, indexdecl);
    use(d, indexdecl) = t;
    if (DEBUG_SR) {
      cout << "AAS::adjoinConstant ";
      cout << PLCODE(t);
      cout << endl;
    }
  }
}

void ArrayAccessSet::adjoin(TreeNode *t, TreeNode *WRTloop)
{
  MIVE *m = t->index()->getMIVE(WRTloop);
  if (m == NULL)
    adjoinConstant(t, WRTloop);
  else {
    if (DEBUG_SR) {
      cout << "AAS::adjoin ";
      cout << PLCODE(t);
      cout << endl;
    }
    MIVEset * & s = MIVEs(t->array());
    if (s == NULL)
      s = new MIVEset;
    ::adjoin(s, m);
    use(arrayVarDecl(t->array())) = t;
  }
}

bool ArrayAccessSet::containsOSR(TreeNode *t, TreeNode *WRTloop)
{
  return (baseMIVEForOSR(t->array(), t->index()->getMIVE(WRTloop)) != NULL);
}

bool ArrayAccessSet::containsConstant(TreeNode *t, TreeNode *WRTloop)
{
  TreeNode *d = arrayVarDecl(t->array());
  TreeNode *indexdecl = t->index()->decl()->source();
  bool result =
    (d != NULL && indexdecl != NULL &&
     (c[d] != NULL || (c.erase(d), false)) &&
     ::contains(c[d], indexdecl));
  if (DEBUG_SR) {
    cout << "AAS::containsConstant ";
    cout << PLCODE(t);
    cout << ": " << result << endl;
  }
  return result;
}

bool ArrayAccessSet::contains(TreeNode *t, TreeNode *WRTloop)
{
  TreeNode *d = arrayVarDecl(t->array());
  if (d == NULL)
    return false;
  const MIVE *im = t->index()->getMIVE(WRTloop);
  bool result =
    (im == NULL && containsConstant(t, WRTloop) ||
     im != NULL &&
     (m[d] != NULL || (m.erase(d), false)) &&
     ::contains(m[d], t->index()->getMIVE(WRTloop)));
  if (im != NULL && DEBUG_SR) {
    cout << "AAS::contains ";
    cout << PLCODE(t);
    cout << ": " << result << endl;
  }
  return result;
}
