#include <fstream.h>
#include "AST.h"
#include "CodeContext.h"
#include "CfHeader.h"
#include "CfSource.h"
#include "aux-code.h"
#include "decls.h"
#include "interface.h"
#include "lgMacro.h"
#include "lower.h"
#include "code.h"



void InterfaceDeclNode::codeGen()
{
  ClassDecl &declaration = *decl();
  const CtType &descriptorType = declaration.cDescriptorType();
  const string  descriptorName = declaration.cDescriptorName();
  const CtType &className = decl()->cType();
  
  // Declarations
  {
    CfHeader os( declaration.cType() );
    
    declaration.includeSupers( os );
    
    declaration.cType().define( os );
    os << "extern " << descriptorType << ' ' << descriptorName << ";\n"
       << "\nextern void " << MANGLE_CLASS_INIT(<<, className) << "();\n";

    foriter (member, members()->allChildren(), ChildIter)
      if ((*member)->decl()->category() == Decl::Method)
	{
	  const MethodDecl &decl = static_cast<const MethodDecl &>(*(*member)->decl());
	  os << "#define IFACE_DISPATCH_" << decl.cMethodNameStatic()
	     << "(instance)  ((" << cMethodTypeString( decl ) << ") "
	     << lgMacro( "IFACE_DISPATCH", decl.modifiers() & Local )
	     << "(instance, " << FindIntfMethodId( decl ) << "))\n";
	}
  }

  // Definitions
  {
    CfSource os( declaration.cType() );
    
    declaration.includeSelf( os );

    const char *listName = declaration.emitInterfaceList( os );
    os << descriptorType << ' ' << descriptorName << " = {\n"
       << "  Interface, " << listName << "\n};\n\n";
    
    emitStaticFields(declaration, staticFieldDecl, os);
    os << "\nvoid " << MANGLE_CLASS_INIT(<<, className) << "()\n";
    {
      CodeContext subcontext( os );
      vector< TreeNode * > inits;

      foriter (member, members()->allChildren(), ChildIter)
	if ((*member)->decl()->modifiers() & Static)
	  inits.push_back( *member );
	
      EncapsulateInits(inits)->emitStatement(subcontext);
    }
  }
}
