#include <iostream>
#include <cstddef>
#include "AST.h"
#include "CtRegistry.h"
#include "CtType.h"
#include "LocalVars.h"
#include "parse.h"


LocalVars::LocalVars( CtRegistry &dependencies )
  : dependencies( dependencies )
{
}


LocalVars::LocalVars( const LocalVars &parent )
  : dependencies( parent.dependencies )
{
}


void LocalVars::depend( const CtType &type )
{
  dependencies.add( type );
}


void LocalVars::declare( const string &name, const CtType &type )
{
  insert( value_type( type, name ) );
  depend( type );
}


void LocalVars::codeGen( ostream &sink,
			 int curr_debug_line,
			 File *curr_debug_file ) const
{
  // It seems like it should be easy, here, to cluster all locals with
  // the same type on the same line, delimited by commas:
  //
  //    Foo x, y, z;
  //
  // Unfortunately, that does not work properly when the type is a
  // pointer:
  //
  //    Foo * x, y, z;
  //
  // Oh well.
  
#if 1
  const_iterator scan = begin();
  while (scan != end())
    {
      const_iterator last = upper_bound( (*scan).first );
      
      sink << (*scan).first << ' ' << (*scan).second;      
      while (++scan != last)
	sink << ", " << (*scan).second;

      sink << ";\n";

      extern bool line_directives;
      if (line_directives && curr_debug_line != -1)
	sink << "#line " << curr_debug_line
	     << " \"" << *(curr_debug_file->name()) << "\"\n";
    }
#else
  for (const_iterator scan = begin(); scan != end(); ++scan) {
    sink << (*scan).first << ' ' << (*scan).second << ";";
    if (curr_debug_line != -1) {
      assert(curr_debug_file != NULL);
      sink << '\n'
	   << "#line "
	   << curr_debug_line
	   << " \""
	   << *(curr_debug_file->name())
	   << "\""
	   << '\n';
    } else {
      sink << '\n';
    }
  }
#endif
}


#ifdef __KCC
void *LocalVars::operator new( size_t )
{
  return 0;
}
#endif
