#ifndef _UNIQUE_CONS_H_
#define _UNIQUE_CONS_H_

#include <map>
#include "ullist.h"

#define generic template <class T>

generic class UniqueCons;

generic class UniqueCons {
  typedef T value_type;
  typedef ullist<T> list_type;
  typedef ullist<T> * list_ref;
  typedef const ullist<T> * const_list_ref;
  typedef map< const_list_ref, const_list_ref > map_list_to_list;
  typedef map< value_type, map_list_to_list > map_value_list_to_list;
  
private:
  map_value_list_to_list m;
  map_value_list_to_list ma;

public:  
  const_list_ref cons(value_type v, const_list_ref l) {
    const_list_ref & result = m[v][l];
    if (result)
      return result;
    else
      return (result = new list_type(v, l));
  }
  const_list_ref cons(value_type v) { return cons(v, NULL); }

  ~UniqueCons() {
    for (map_value_list_to_list::const_iterator i = m.begin();
	 i != m.end();
	 i++)
      for (map_list_to_list::const_iterator j = (*i).second.begin();
	   j != (*i).second.end();
	   j++)
	delete (*j).second;
  }

  const_list_ref append(const_list_ref L0, const_list_ref L1) {
    if (L0 == NULL)
      return L1;
    else if (L1 == NULL)
      return L0;
    else 
      return cons(L0->front(), append(L0->tail(), L1));
  }

  const_list_ref reverse (const_list_ref L)
  {
    const_list_ref result = NULL;
    if (L == NULL)
      return NULL;
    foreach_const (p, list_type, *L)
      result = cons (*p, result);
    return result;
  }

  /* Set operations on ordered lists */

  const_list_ref adjoin(value_type v, const_list_ref l)
  {
    const_list_ref & result = ma[v][l];
    if (result)
      return result;
    else if (l == NULL)
      return (result = cons(v));
    else if (l->front() == v)
      return (result = l);
    else if (l->front() > v)
      return (result = cons(v, l));
    else
      return (result = cons(l->front(), adjoin(v, l->tail())));
  }
  
  const_list_ref merge(const_list_ref l0, const_list_ref l1)
  {
    if (l0 == NULL)
      return l1;
    else if (l1 == NULL || l1 == l0)
      return l0;
    else if (l0->front() < l1->front())
      return cons(l0->front(), merge(l0->tail(), l1));
    else if (l0->front() > l1->front())
      return cons(l1->front(), merge(l1->tail(), l0));
    else
      return cons(l0->front(), merge(l1->tail(), l0->tail()));
  }

  const_list_ref intersect(const_list_ref l0, const_list_ref l1)
  {
    if (l0 == NULL || l1 == NULL)
      return NULL;
    else if (l0 == l1)
      return l0;
    else if (l0->front() < l1->front())
      return intersect(l0->tail(), l1);
    else if (l0->front() > l1->front())
      return intersect(l1->tail(), l0);
    else
      return cons(l0->front(), intersect(l1->tail(), l0->tail()));
  }
  
  size_t size(const_list_ref l0) {
    return l0 ? l0->size() : 0;
  }

};

generic ullist<T>* find(T x, ullist<T>* l)
{
  while (l)
    if (l->front() == x) return l;
    else l = l->tail();

  return NULL;
}

generic bool find(bool f(T x), ullist<T> *l)
{
  while (l != NULL)
    if (f(l->front()))
      return true;
    else
      l = l->tail();

  return false;
}

#undef generic

#endif _UNIQUE_CONS_H_
