#include "AST.h"
#include "CodeContext.h"
#include "CfSource.h"
#include "CtType.h"
#include "MethodDecl.h"
#include "code.h"
#include "code-assign.h"
#include "decls.h"
#include "optimize.h"
#include "lower.h"
#include "Poly.h"

void FieldDeclNode::codeGen( CodeContext &context )
{
  TypeNode *itype;
  string initializer; 
  if (initExpr()->absent()) {
    initExpr(defaultInitializer(dtype(), position()));
    itype = dtype();
  } else
    itype = initExpr()->type();

  if (initializer.empty())
    initializer = initExpr()->emitExpression( context );

  if ((decl()->modifiers() & Static))
    context << makeAssign("&STATIC_REF(" + cFieldName(decl()) + ")",
			  initializer, true, assignKind(itype), true);
  else
    context << MANGLE_THIS_VAR(<<) << "->" << cFieldName(decl()) << " = " << initializer;

  context << ';' << endCline;
}

void VarDeclNode::codeGen( CodeContext &context )
{
  const string varName =
    ((decl()->modifiers() & Common::CompilerGenerated) ?
     MANGLE_TEMP(+, *simpName()->simpName()->ident()) :
     MANGLE_STACK_VAR(+, *simpName()->simpName()->ident()));

  context.declare( varName, dtype()->cType() );
}


void MethodDeclNode::codeGen( CfSource &os )
{
  MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );

  if (!(mdecl.modifiers() & (Common::Native | Common::Abstract))) {
    bool first = true;
    extern bool line_directives;

    if ((debugger_level > 0) && (line_directives)) {
      const SourcePosn pos = position();
      if (!pos.unknown()) {
	os << "#line " << pos.posnToLineNumber()
	   << " \"" << *(pos.file->name()) << "\"\n";
      }
    }

    os << returnType()->cType() << ' '
       << mdecl.cMethodNameStatic() << '(';
    
    const TypeNode *thisType = mdecl.thisType();
    if (thisType) {
      os << thisType->cType() << " " << MANGLE_THIS_VAR(<<);
      first = false;
    }	

    foriter (param, params()->allChildren(), TreeNode::ChildIter) {
      if (first) first = false; else os << ",";
      os << (*param)->dtype()->cType()
	 << " " << MANGLE_STACK_VAR(<<, *(*param)->simpName()->ident());
    }

    os << ")";
    {
      CodeContext context(os);
      // cout << "codeGen: " << mdecl.cMethodNameStatic() << endl;
      body()->codeGen(context);
    }

    os << '\n';

    freePolys();
  }
}


void MethodSignatureNode::codeGen ( CodeContext & )
{
}


void ConstructorDeclNode::codeGen( CfSource &os )
{
  const MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );
  const TypeNode &thisType = *mdecl.thisType();
  const string thisTypeName = thisType.cType();
  extern bool line_directives;

  if ((debugger_level > 0) && (line_directives)) {
    const SourcePosn pos = position();
    if (!pos.unknown()) {
      os << "#line " << pos.posnToLineNumber()
	 << " \"" << *(pos.file->name()) << "\"\n";
    }
  }
  
  os << thisTypeName << ' '
     << decl()->cMethodNameStatic()
     << "(";

  bool first = true;
  if (!thisType.isImmutable()) {
    os << thisTypeName << " " << MANGLE_THIS_VAR(<<);
    first = false;
  }

  foriter (param, params()->allChildren(), TreeNode::ChildIter) {
    if (!first) os << ",";
    os << (*param)->dtype()->cType()
       << " " << MANGLE_STACK_VAR(<<, *(*param)->simpName()->ident());
    first = false;
  }
  os << ")";

  {
    CodeContext context( os );
    const MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );
    const TypeNode &thisType = *mdecl.thisType();
    
    if (thisType.isImmutable()) {
      // declare me
      context.declare( MANGLE_THIS_VAR(+), thisType.cType() );
      // zero out the contents of me
      context << "memset( &" << MANGLE_THIS_VAR(<<) << ", 0, sizeof(" << MANGLE_THIS_VAR(<<) << ") );" << endCline
	      << decl()->container()->cInitFieldsName()
	      << "( &" << MANGLE_THIS_VAR(<<) << " );" << endCline;
    } else {
      constructorCall()->codeGen(context);
      context << endCline;
    }
    
    body()->codeGen(context);
    context << "return " << MANGLE_THIS_VAR(<<) << ";" << endCline;
  }
  os << "\n\n";
}
