#include "AST.h"
#include "MethodSet.h"
#include "context.h"
#include "decls.h"
#include "solver.h"


void ConstructorDeclNode::inferLocal( const LocalContext &context )
{
  if (context.isExported)
    decl()->type()->paramTypes()->localConservative();
}


void DataDeclNode::inferLocal( const LocalContext & )
{
  dtype()->localProtect();
  
  if (!initExpr()->absent())
    initExpr()->type()->localWeaklySubsumedBy( *dtype() );
}


void FieldDeclNode::inferLocal( const LocalContext &context )
{
  if (context.isExported)
    dtype()->localConservative();

  DataDeclNode::inferLocal( context );
}


void MethodNode::inferLocal( const LocalContext &context )
{
  MethodDecl &myDecl = static_cast< MethodDecl & >( *decl() );
  assert( myDecl.category() == Decl::Method );
  const TypeNode &myType = *myDecl.type();

  if (context.isExported)
    myType.localConservative();

  {
    returnType()->localProtect();
    
    const TypeNode * const thisType = myDecl.thisType();    
    if (thisType && thisType->localNeedsProtection())
      solver().conservative( myType );
  }

  {
    const Decl * const overrides = myDecl.overrides();
    if (overrides)
      myType.localIdenticalTo( *overrides->type() );
  }

  {
    const MethodSet &implements = myDecl.implements();
    for (MethodSet::const_iterator interface = implements.begin();
	 interface != implements.end(); ++interface)
      myType.localIdenticalTo( *(*interface)->type() );
  }
}


void MethodDeclNode::inferLocal( const LocalContext &context )
{
  MethodNode::inferLocal( context );

  if (decl()->modifiers() & Native || isMain())
    decl()->type()->localConservative();
}


void ParameterNode::inferLocal( const LocalContext & )
{
  dtype()->localProtect();
}
