#include "AST.h"
#include "decls.h"
#include "enact-context.h"
#include "stats.h"


void TreeNode::localEnactChildren( const LocalEnactContext &context ) const
{
  for (unsigned sweep = arity(); sweep--; )
    child( sweep )->localEnactTree( context );
}


void TreeNode::localEnactTree( const LocalEnactContext &context )
{
  localEnact( context );
  localEnactChildren( context );
}


////////////////////////////////////////////////////////////////////////


void CompileUnitNode::localEnactChildren( const LocalEnactContext &context ) const
{
  PackageDecl * const decl = static_cast< PackageDecl * >( thePackage );
  assert( decl->category() == Decl::Package );

  const LocalEnactContext subcontext( context, decl );
  TreeNode::localEnactChildren( subcontext );
}
  

void ConstructorDeclNode::localEnactChildren( const LocalEnactContext &context ) const
{
  const bool thisLocal = !(decl()->container()->modifiers() & Immutable);
  const LocalEnactContext subcontext( context, thisLocal );
  TreeNode::localEnactChildren( subcontext );
}


void FieldDeclNode::localEnactChildren( const LocalEnactContext &context ) const
{
  const bool thisLocal = !(decl()->container()->modifiers() & Immutable);
  const LocalEnactContext subcontext( context, thisLocal );
  TreeNode::localEnactChildren( subcontext );
}


void MethodDeclNode::localEnactChildren( const LocalEnactContext &context ) const
{
  // !! potential problem: disagreement between decl modifiers and
  // !! decl type modifiers
  
  const bool thisLocal = decl()->type()->isLocal();
  const LocalEnactContext subcontext( context, thisLocal );
  TreeNode::localEnactChildren( subcontext );
}
