#include "AST.h"
#include "context.h"
#include "decls.h"


void TreeNode::inferLocalTree( const LocalContext &context )
{
  inferLocalChildren( context );
  inferLocal( context );
}


void TreeNode::inferLocalTreeExport( const LocalContext &context )
{
  const Decl &declaration = *decl();
  const Modifiers mods = declaration.modifiers();
  const bool isExported = mods & (Public | Protected);

  const LocalContext subcontext( context, isExported );
  TreeNode::inferLocalTree( subcontext );
}


////////////////////////////////////////////////////////////////////////


void ClassDeclNode::inferLocalTree( const LocalContext &context )
{
  inferLocalTreeExport( context );
}


void CompileUnitNode::inferLocalTree( const LocalContext &context )
{
  const bool isExported = !(codeGen_main && selectedForCodeGen( false ));
  const LocalContext subcontext( context, isExported );
  TreeNode::inferLocalTree( subcontext );
}


void ConstructorDeclNode::inferLocalTree( const LocalContext &context )
{
  inferLocalTreeExport( context );
}


void FieldDeclNode::inferLocalTree( const LocalContext &context )
{
  inferLocalTreeExport( context );
}


void InterfaceDeclNode::inferLocalTree( const LocalContext &context )
{
  inferLocalTreeExport( context );
}


void MethodNode::inferLocalTree( const LocalContext &context )
{
  inferLocalTreeExport( context );
}
