#ifndef _INCLUDE_TITANIUM_OPTIMIZE_H_
#define _INCLUDE_TITANIUM_OPTIMIZE_H_

extern bool opt_lift;
extern bool opt_sr;
extern bool opt_osr;
extern bool opt_dce;

#include "AST.h"
#include "code-util.h"
#include "cfg.h"

/* too verbose for most purposes */
#define DEBUG_DOM_VERBOSE (DEBUG_PHASE_ENABLED("domv"))

/* dumpphase opt turns on lots of debugging stuff. */
#define DEBUG_OPT DEBUG_PHASE_ENABLED("opt")
#define DEBUG_SR (DEBUG_OPT || DEBUG_PHASE_ENABLED("sr"))
#define DEBUG_OSR (DEBUG_OPT || DEBUG_PHASE_ENABLED("osr"))
#define DEBUG_DCE (DEBUG_OPT || DEBUG_PHASE_ENABLED("dce"))
#define DEBUG_LOOP_INVAR (DEBUG_OPT || DEBUG_PHASE_ENABLED("invar"))
#define DEBUG_LOOP_ANAL (DEBUG_OPT || DEBUG_PHASE_ENABLED("loopanal"))
#define DEBUG_LIFT (DEBUG_OPT || DEBUG_PHASE_ENABLED("lift"))
#define DEBUG_DOM \
    (DEBUG_DOM_VERBOSE || DEBUG_OPT || DEBUG_PHASE_ENABLED("dom"))

#define LOOP_ARITY(l) ((l)->vars()->child(0)->initExpr()->type()->tiArity())

#include "Worklist.h"

TreeNode *CloneTree(TreeNode *t, bool cleanUseDef = false);
TreeNode *newBoolLit(bool b, SourcePosn p = NoSourcePosition);

treeSet *allDefs(treeSet *vars);
treeSet *allUses(TreeNode *t);
treeSet *LvalueUses(TreeNode *t);
treeSet *RvalueUses(TreeNode *t);
treeSet *ObjectNodes(TreeNode *e);
treeSet *treeSetAdjoin(treeSet *s, const TreeNode *t);
treeSet *adjoin(treeSet *s, const TreeNode *t);
treeSet *destructively_intersect(treeSet *s, treeSet *t);
treeSet *destructively_union(treeSet *s, treeSet *t);
treeSet *copy(treeSet *s);

map_tree_to_treeSet *
    destructive_convert_cfgmap_to_astmap(map_CFGnode_to_CFGset *r);
treeSet *
    destructive_convert_cfgset_to_treeset(CFGset *s);
map_tree_to_treeSet *
    destructive_convert_cfgmap_to_astmap(map_CFGnode_to_CFGuset *r);
treeSet *
    convert_cfguset_to_treeset(const CFGuset *s);

treeSet *exprset_to_stmtset(treeSet *t);

bool isSingletonContaining(TreeNode *e, treeSet *s);

CFGset *destructively_intersect(CFGset *s, CFGset *t);
bool contains(const CFGset *s, const CfgNode *t);
CFGset *adjoin(CFGset *s, const CfgNode *t);

bool set_contains_any(treeSet *s, llist<TreeNode *> *l);
bool noDefsFromSet(TreeNode *t, treeSet *s);

void initInvar(TreeNode *l);
void foundInvariant(TreeNode *inv, TreeNode *WRTloop);
bool isInvar(const TreeNode *t, TreeNode *WRTloop);

TreeNode *valueTransferred(TreeNode *def, TreeNode *use);

void propagateChanges(Worklist *w, TreeNode *t);

void lazyOptimize();

treeSet * list_to_set(llist<TreeNode *> *l);
llist<TreeNode *> * set_to_list(const treeSet *s);
bool isSubset(const treeSet &s, const treeSet &t);
bool isSubset(const treeSet *s, const treeSet *t);
bool treeSetContains(const treeSet *s, const TreeNode *t);
bool treeSetContains(const treeSet &s, const TreeNode *t);
bool contains(const treeSet *s, const TreeNode *t);
bool contains(const treeSet &s, const TreeNode *t);
void printTreeSet(const treeSet *u, ostream &os, int indent);
void printTreeSetWithParents(const treeSet *s, ostream &os, int m);
treeSet *filter(bool f(TreeNode *t), const treeSet *s);

BlockNode *SortVarDeclsInBlock(BlockNode *b);
SwitchBranchNode *SortVarDeclsInSwitch(SwitchBranchNode *b);

// TreeNode *removeUselessVarDecls(TreeNode *t, treeSet *useless);

#define PCODE2(os, t) ((os).flush(), (t)->pseudoprint((os), 0), "")
#define PLCODE2(os, t) ((os).flush(),					\
			((os) << "{" << (t)->oper_name() << " at " <<	\
			 (t)->position().asString() << "} "),		\
			(t)->pseudoprint((os), 0),			\
			"")
#define POPER2(os, t) (((os) << "{" << (t)->oper_name() << " at " <<	\
			(t)->position().asString() << "}"),		\
		       "")
#define SHOWMIVE2(os, m) ((m)->print((os)), "")

#define PCODE(t) PCODE2(cout, t)
#define SAFEPCODE(t) (((t) == NULL) ? "NULL" : PCODE(t))
#define SAFEPLCODE2(os, t) (((t) == NULL) ? "NULL" : PLCODE2(os, t))
#define POPER(t) POPER2(cout, t)
#define PLCODE(t) PLCODE2(cout, t)
#define SHOWMIVE(t) SHOWMIVE2(cout, t)

#define QCFG(x) (unique((void *) (x)))

#endif /* _INCLUDE_TITANIUM_OPTIMIZE_H_ */
