#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include "lapi_comm.h"

char *RUNTIME_MODEL = "sp2clump";

extern void monitor_system_init();

int use_mem_hack = 1;
int no_bulk_copies = 0; 
swclock_handle_t profclock; /* For SP/2 high-res timing. */
static pthread_t *threads;  /* Stores all the thread handles. */
pthread_key_t myProcKey;    /* For thread-local data. */
int THREADS_PER_NODE = 0;  
int ti_argc; char **ti_argv;  /* Global versions of the args. */
int *node_map;              /* Maps threads to nodes. */
int THREADS_DONE = 0;
pthread_mutex_t global_lock;

void *thr_main(void *id) {
  pthread_setspecific(myProcKey, id);
  if (bindprocessor(BINDTHREAD, thread_self(), MYTHREAD + 1) < 0) {
    perror("Can't give thread its own CPU - continuing execution");
  }
  STATIC_REF(BroadcastCycle) = 0;
  ti_main(ti_argc, ti_argv);
  NODELOCK;
  THREADS_DONE++;
  NODEUNLOCK;
  return NULL;
}

void spawn(void) {
  int count;
  void *thr_retval;
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_mutex_init(&global_lock, NULL);

  threads = (pthread_t *)malloc(THREADS_PER_NODE * sizeof(pthread_t));
  PROCS = THREADS_PER_NODE * NODES;
  printf("Spawning %d threads on %d nodes.\n", THREADS_PER_NODE, NODES);
  barrier_init();
  node_map = (int *)malloc(PROCS * sizeof(int));
  pthread_key_create(&myProcKey, NULL);
  for (count = 0; count < PROCS; count++) {
    node_map[count] = count / THREADS_PER_NODE;
  }
  pthread_attr_setstacksize(&attr, 2000000);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
  for (count = 0; count < THREADS_PER_NODE; count++) {
    pthread_create(&threads[count], &attr, thr_main, (void *)(MYNODE * THREADS_PER_NODE + count));
  }
  for (count = 0; count < THREADS_PER_NODE; count++) {
    pthread_join(threads[count], &thr_retval);
  }
}

/* Figure out if the -threads switch is there. */
void processArgs(int *argcP, char ***argvP) {
  if (strcmp((*argvP)[1], "-threads") == 0) {
    THREADS_PER_NODE = atoi((*argvP)[2]);
    (*argcP) -= 2;
    (*argvP) += 2;
  }
  else {
    /* Use the system reported max number of CPUs. */
    THREADS_PER_NODE = sysconf(_SC_NPROCESSORS_ONLN);
  }
  if (!THREADS_PER_NODE) {
    printf("Error parsing -threads parameter.\n");
    exit(-1);
  }
}
  
int main(int argc, char **argv) {
  int rcode;

  region_init();

  if (rcode = sp2_init()) {
    fprintf(stderr, "Failed to initialize SP/2 communications layer.\n");
    exit(rcode);
  }

  profclock = swclockInit();
  processArgs(&argc, &argv);
  monitor_system_init(); /* Depends on THREADS_PER_NODE */
  ti_argc = argc; ti_argv = argv;
  spawn();

  sp2_shutdown();
}
