#ifndef _INCLUDE_MEM_LOCAL_H_
#define _INCLUDE_MEM_LOCAL_H_


struct class_header;


/* Assign coding:
   ASSIGN is used for fields and array references, not for local vars
   [WEAK_][S][PL|PG]ASSIGN_[LOCAL|GLOBAL]:
   S: destination is a static field
   LOCAL: destination is accessed by a local pointer
   GLOBAL: destination is accessed by a global pointer
   PL: value is a local reference
   PG: value is a global reference
   WEAK_: "weak" assignment

   The arguments to assign must be evaluable multiple times.
   */   

#define DEREF_LOCAL(val, ptr)			do { (val) = *(ptr); } while (0)
#define ASSIGN_LOCAL(ptr, val)			do { *(ptr) = (val); } while (0)

#ifdef NOMEMCHECKS

#define PLASSIGN_LOCAL(ptr, val)		(*(ptr) = (val))
#define PGASSIGN_LOCAL(ptr, val)		(*(ptr) = (val))
#define SPLASSIGN_LOCAL(ptr, val)		(*(ptr) = (val))
#define SPGASSIGN_LOCAL(ptr, val)		(*(ptr) = (val))

#else

#ifdef ENABLE_RC

#define RC_PLASSIGN_LOCAL(ptr, ptrRegion, val, valRegion) \
    { \
      RegionId oldValRegion; \
      PLSWAP_LOCAL(ptr, val, oldValRegion); \
      if (oldValRegion != valRegion) { \
	  if (oldValRegion != ptrRegion) oldValRegion->rc--; \
	  if (valRegion != ptrRegion) valRegion->rc++; \
	} \
    }

#define RC_PGASSIGN_LOCAL(ptr, ptrRegion, val, valRegion) \
    { \
      RegionId oldValRegion; \
      PGSWAP_LOCAL(ptr, val, oldValRegion); \
      if (oldValRegion != valRegion) { \
	  if (oldValRegion != ptrRegion) oldValRegion->rc--; \
	  if (valRegion != ptrRegion) valRegion->rc++; \
	} \
    }

#define SPLASSIGN_LOCAL(ptr, val) \
  do { \
    RegionId oldValRegion = PL2RegionId(*(ptr)); \
    RegionId valRegion = PL2RegionId(val); \
    if (oldValRegion != valRegion) \
      { \
	oldValRegion->rc--; \
	valRegion->rc++; \
      } \
    *(ptr) = val; \
  } while (0)

#define SPGASSIGN_LOCAL(ptr, val) \
  do { \
    RegionId oldValRegion = PG2RegionId(*(ptr)); \
    RegionId valRegion = PG2RegionId(val); \
    if (oldValRegion != valRegion) \
      { \
	oldValRegion->rc--; \
	valRegion->rc++; \
      } \
    *(ptr) = val; \
  } while (0)

#else

#define RC_PLASSIGN_LOCAL(ptr, ptrRegion, val, valRegion) (*(ptr) = (val))
#define RC_PGASSIGN_LOCAL(ptr, ptrRegion, val, valRegion) (*(ptr) = (val))
#define SPLASSIGN_LOCAL(ptr, val) (*(ptr) = (val))
#define SPGASSIGN_LOCAL(ptr, val) (*(ptr) = (val))

#endif

#define PLASSIGN_LOCAL(ptr, val) \
  do { \
    RegionId ptrRegion = PL2RegionId(ptr); \
    RegionId valRegion = PL2RegionId(val); \
    if (!SHARED_REGIONID(valRegion) && SHARED_REGIONID(ptrRegion)) sharingViolation(); \
    RC_PLASSIGN_LOCAL(ptr, ptrRegion, val, valRegion); \
  } while (0)

#define PGASSIGN_LOCAL(ptr, val) \
  do { \
    RegionId ptrRegion = PL2RegionId(ptr); \
    RegionId valRegion = PG2RegionId(val); \
    if (!SHARED_REGIONID(valRegion) && SHARED_REGIONID(ptrRegion)) sharingViolation(); \
    RC_PGASSIGN_LOCAL(ptr, ptrRegion, val, valRegion); \
  } while (0)



#endif

#define WEAK_ASSIGN_LOCAL			ASSIGN_LOCAL
#define WEAK_PLASSIGN_LOCAL			PLASSIGN_LOCAL
#define WEAK_PGASSIGN_LOCAL			PGASSIGN_LOCAL
#define WEAK_SPLASSIGN_LOCAL			SPLASSIGN_LOCAL
#define WEAK_SPGASSIGN_LOCAL			SPGASSIGN_LOCAL

#define FIELD_ADDR_LOCAL(addr, ptr, fld)	((addr) = &(ptr)->fld)
#define IFACE_DISPATCH_LOCAL(ptr, mid)		(ptr->intf_info[mid])
#define INDEX_LOCAL(element, base, offset)	(element) = (base) + (offset)
#define SUM_LOCAL(type, element, base, offset)	(element) = (type)(((char *)(base)) + (offset))
#define EQUAL_LOCAL(p, q)			((p) == (q))

#define CLASS_INFO_LOCAL(val, type, obj)	((val) = (obj)->class_info)

#define PRE_INCR_LOCAL( val, ptr)		((val) = ++*(ptr))
#define PRE_DECR_LOCAL( val, ptr)		((val) = --*(ptr))
#define POST_INCR_LOCAL(val, ptr)		((val) = (*(ptr))++)
#define POST_DECR_LOCAL(val, ptr)		((val) = (*(ptr))--)


#endif /* !_INCLUDE_MEM_LOCAL_H_ */
