#include <limits.h>
#include <pwd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

#include <class_header.h>
#include <java_string.h>
#include <native-stubs.h>
#include "runtime-options.h"
#include "layout!PTAjbyte.h"
#include "layout!PLti5cdescmT6Object4lang4java.h"
#include "layout!Pclass_header.h"
#include "layout!Pjbyte.h"
#include "layout!Pjint.h"
#include "layout!TAjbyte.h"

extern char *RUNTIME_MODEL;

static void checkIsArray( PT6Object4lang4java object )
{
    typedef GP_type( TypeCategory ) GP_TypeCategory;

    ti5cdescmT6Object4lang4java *info;
    Pclass_header header;
    GP_TypeCategory categoryField;
    TypeCategory category;

    CLASS_INFO_GLOBAL( info, ti5cdescmT6Object4lang4java, object );
    TO_GLOBAL( header, TO_PROC( object ), (class_header *) info );
    FIELD_ADDR_GLOBAL( categoryField, header, category );
    DEREF_GLOBAL( category, categoryField );

    assert( category == JavaArray );
}
  

void m9arraycopyPT6Object4lang4javaIPT6Object4lang4javaIImT6System4lang4java(
    PT6Object4lang4java src,
    jint src_pos,
    PT6Object4lang4java dst,
    jint dst_pos,
    jint length)
{
    PTAjbyte src_array, dst_array;
    jint src_size, dst_size;
    Pjbyte src_data, dst_data;
    checkIsArray( src );
    checkIsArray( dst );

    src_array = * (PTAjbyte *) &src;
    dst_array = * (PTAjbyte *) &dst;

    JAVA_ARRAY_SIZE_GLOBAL( src_size, src_array );
    JAVA_ARRAY_SIZE_GLOBAL( dst_size, dst_array );
    assert( src_size == dst_size );

    JAVA_ARRAY_CHECK_GLOBAL( src_array, src_pos, length );
    JAVA_ARRAY_CHECK_GLOBAL( dst_array, dst_pos, length );

    /* We would like to use JAVA_ARRAY_ADDR_GLOBAL for the
       indexing computations below, but we cannot.  The arrays
       might not really be arrays of bytes, and so the bounds
       check inside JAVA_ARRAY_ADDR_GLOBAL will fail
       inappropriately.
       */
    
    FIELD_ADDR_GLOBAL( src_data, src_array, data[0] );
    INDEX_GLOBAL( src_data, src_data, src_pos * src_size );
    
    FIELD_ADDR_GLOBAL( dst_data, dst_array, data[0] );
    INDEX_GLOBAL( dst_data, dst_data, dst_pos * dst_size );

    {
	  int copy_sz = src_size * length;
    int islocalsrc = isLocal(src_data);
    int islocaldest = isLocal(dst_data);

    if (islocalsrc && islocaldest) {
      memcpy(TO_LOCAL(dst_data), TO_LOCAL(src_data), copy_sz);
      }
    else if (islocalsrc && !islocaldest) {
	    local_to_global_copy(TO_LOCAL(src_data), dst_data, copy_sz);
      }
    else if (!islocalsrc && islocaldest) {
	    global_to_local_copy(src_data, TO_LOCAL(dst_data), copy_sz);
      }
    else { /* !islocalsrc && !islocaldest */
	    VARARRAY_DECLARE(jbyte, tmp_buf, copy_sz);
	    VARARRAY_CREATE(jbyte, tmp_buf, copy_sz);
	    global_to_local_copy(src_data, tmp_buf, copy_sz);
	    local_to_global_copy(tmp_buf, dst_data, copy_sz);
	    VARARRAY_DESTROY(tmp_buf);
      }
    }
}


jlong m17currentTimeMillismT6System4lang4java()
{
    struct timeval now;
    assert(!gettimeofday(&now, 0));
    
    /* We could round the secs, but it does not seem worth it. */
    return now.tv_sec * (jlong) 1000 + now.tv_usec / 1000;
}

static void addProperty( PT10Properties4util4java props, char keyC[], char valueC[] )
{
  if (valueC)
    {
      const LP_JString key = java_string_build_8( keyC );
      const LP_JString value = java_string_build_8( valueC );
      m11addPropertyPT10Properties4util4javaLT6String4lang4javaLT6String4lang4javamT11NativeUtils4lang2ti( props, key, value );
    }
}    
    

PT10Properties4util4java
m14initPropertiesPT10Properties4util4javamT6System4lang4java( PT10Properties4util4java props )
{
  char * const cwd = getcwd( 0, PATH_MAX );
  const struct passwd * const user = getpwuid( getuid() );

  addProperty( props, "java.version",		"0" );
  addProperty( props, "java.vendor",		"Titanium Project" );
  addProperty( props, "java.vendor.url",	"http://HTTP.CS.Berkeley.EDU/Research/Projects/titanium/" );
  addProperty( props, "java.home",		"/" );
  addProperty( props, "java.class.version",	"0" );
  addProperty( props, "java.class.path",	"/" );
  addProperty( props, "os.name",		"Titanium" );
  addProperty( props, "os.arch",		"Titanium" );
  addProperty( props, "os.version",		"0" );
  addProperty( props, "file.separator",	"/" );
  addProperty( props, "path.separator",	":" );
  addProperty( props, "line.separator",	"\n" );
  /* Possible values: split-c, smp, uniproc, pthread */
  addProperty( props, "runtime.model", RUNTIME_MODEL);
  if ((strcmp(RUNTIME_MODEL, "split-c") == 0) ||
      (strcmp(RUNTIME_MODEL, "sp2") == 0) ||
      (strcmp(RUNTIME_MODEL, "cray-t3e") == 0)) { 
    addProperty( props, "runtime.distributed", "true");
  }
  else {
    addProperty( props, "runtime.distributed", "false");
  }

  if (user) {
    addProperty( props, "user.name", user->pw_name );
    addProperty( props, "user.home", user->pw_dir );
  }
    
  if (cwd)
    addProperty( props, "user.dir", cwd );

  return props;
}
