#ifndef TITANIUM_GC_H
#define TITANIUM_GC_H

#include "config.h"

#ifdef USE_GC_NONE

#  define ti_malloc(s)			calloc(1, s)
#  define ti_malloc_atomic(s)		calloc(1, s)
#  define ti_malloc_huge(s)		calloc(1, s)
#  define ti_malloc_atomic_huge(s)	calloc(1, s)
#  define ti_mem_thr_init(x)		((void) 0)
#  define ti_region_push_all(x,y)	((void) 0)
#  define ti_free			free
#  define ti_region_get_mem		nogc_region_get_mem

#include <stdlib.h>

/* in runtime/pages.c */
extern void *ti_region_get_mem(size_t s);

#else /* USE_GC_NONE */

/* Use Boehm-Weiser GC, at the very least to manage memory */

#  include "backend.h"

/* This is the call the Boehm-Weiser GC uses to request memory from
   us. Size is a multiple of GC_page_size, and we must return something
   that is at an address that is a multiple of GC_page_size */
char *get_memory_from_titanium(unsigned long size);

char *GC_unix_get_mem(unsigned long size);
#define ti_region_get_mem		GC_unix_get_mem

#  ifdef MEMORY_DISTRIBUTED

/* The following lives in runtime/pages.c */
extern void *ti_dist_malloc(size_t);

#    define ti_malloc			ti_dist_malloc
#    define ti_malloc_atomic		GC_malloc_atomic_uncollectable
#    define ti_malloc_huge		ti_dist_malloc
#    define ti_malloc_atomic_huge	GC_malloc_atomic_uncollectable
#    define ti_mem_thr_init(x)		\
       do { \
         GC_stackbottom = (ptr_t)(x); \
	 GC_malloc(1); \
       } while(0)
#    define ti_region_push_all		GC_push_all
#    define ti_free			GC_free

#  else /* MEMORY_DISTRIBUTED */

/*
 * Note: when allocating memory using the "huge" allocators, you must
 * ensure that there will be a live pointer to within the first 256
 * bytes of the block of memory for the duration of its lifetime.
 */

#    define ti_malloc			GC_malloc
#    define ti_malloc_atomic		GC_malloc_atomic
#    define ti_malloc_huge		GC_malloc_ignore_off_page
#    define ti_malloc_atomic_huge	GC_malloc_atomic_ignore_off_page
#    define ti_mem_thr_init(x)		\
       do { \
         GC_stackbottom = (ptr_t)(x); \
	 GC_malloc(1); \
       } while(0)
#    define ti_region_push_all		GC_push_all
#    define ti_free			GC_free

#  endif /* MEMORY_DISTRIBUTED */

#  define ti_region_get_mem		GC_unix_get_mem

#  undef TRUE
#  undef FALSE
#  include <gc/gc.h>
#  include <gc/gc_private.h>

/* This is the call the Boehm-Weiser GC uses to request memory from
   us. Size is a multiple of GC_page_size, and we must return something
   that is at an address that is a multiple of GC_page_size */
/* In runtime/pages.c */
char *get_titanium_memory(GC_word bytes);

ptr_t GC_unix_get_mem(word);

#endif /* USE_GC_NONE */

#endif /* !TITANIUM_GC_H */
