#ifndef INCLUDE_SSTREAM_H
#define INCLUDE_SSTREAM_H

#include "config.h"
#include "using-std.h"


#ifdef HAVE_SSTREAM
#include <sstream>
#else // no stringstream classes

#include <string>
#include <strstream>

// This is not intended to be a complete stringstream implementation.
// It has only the minimal functionality required for use in the
// Titanium compiler, and can be removed once real implementations are
// more widely available.

class stringstream : public strstream {
public:
  stringstream();
  stringstream( const string & );
  string str() const;
  streambuf *rdbuf() const;
  
private:
  string extract();
};


inline stringstream::stringstream()
{
}


inline string stringstream::str() const
{
  return const_cast< stringstream * >( this )->extract();
}


inline streambuf *stringstream::rdbuf() const
{
  return ios::rdbuf();
}


inline string stringstream::extract()
{
  const int wasFrozen = frozen();
  const string result( strstream::str(), pcount() );
  
  freeze( wasFrozen );
  return result;
}


#endif // no stringstream classes


#endif // !INCLUDE_SSTREAM_H
