#include "AST.h"
#include "code-util.h"
#include "compiler.h"
#include "config.h"
#include "decls.h"


void TreeNode::resolveRequires( ClassDecl *currentClass )
{
  foriter (child, allChildren(), ChildIter)
    (*child)->resolveRequires( currentClass );
}


void TypeDeclNode::resolveRequires( ClassDecl * )
{
  TreeNode::resolveRequires( decl() );
}


void MethodCallNode::resolveRequires( ClassDecl *currentClass )
{
  currentClass->requires( *method()->accessedObjectType() );
  TreeNode::resolveRequires( currentClass );
}


void AllocateNode::resolveRequires( ClassDecl *currentClass )
{
  currentClass->requires( *dtype() );
  TreeNode::resolveRequires( currentClass );
}


void AllocateArrayNode::resolveRequires( ClassDecl *currentClass )
{
  TypeNode *descend = type();
  while (descend->isArrayType())
    {
      if (!descend->isJavaArrayType())
	currentClass->requires( *descend );
      
      descend = descend->elementType();
    }
  
  TreeNode::resolveRequires( currentClass );
}


void CatchNode::resolveRequires( ClassDecl *currentClass )
{
  currentClass->requires( *param()->dtype() );
  TreeNode::resolveRequires( currentClass );
}


void DynamicTypeNode::resolveRequires( ClassDecl *currentClass )
{
  if (!dtype()->isPrimitive())
    currentClass->requires( *NativeUtilsDecl );

  if (!(dtype()->isJavaArrayType() || dtype()->isPrimitive()))
    currentClass->requires( *dtype() );
  
  TreeNode::resolveRequires( currentClass );
}


void FieldAccessNode::resolveRequires( ClassDecl *currentClass )
{
  if (decl()->category() == Decl::Field && decl()->modifiers() & Static)
    currentClass->requires( *accessedObjectType() );
  
  TreeNode::resolveRequires( currentClass );
}


void FieldDeclNode::resolveRequires( ClassDecl *currentClass )
{
  // this should really be taken care of by having the rewrite stage
  // add an explicit initializer
  
  if (initExpr()->absent() && dtype()->isImmutable())
    currentClass->requires( *dtype() );

  TreeNode::resolveRequires( currentClass );
}


void VarDeclNode::resolveRequires( ClassDecl *currentClass )
{
  // this should really be taken care of by having the rewrite stage
  // add an explicit initializer
  
  if (initExpr()->absent() && dtype()->isImmutable())
    currentClass->requires( *dtype() );
  
  TreeNode::resolveRequires( currentClass );
}


void DomainNode::resolveRequires( ClassDecl *currentClass )
{
  // this should really be taken care of by having the rewrite stage
  // add an explicit allocation / constructor invocation
  
  currentClass->requires( *type() );
  TreeNode::resolveRequires( currentClass );
}
  

void TitaniumArrayAccessNode::resolveRequires( ClassDecl *currentClass )
{
  currentClass->requires( *array()->type() );
  TreeNode::resolveRequires( currentClass );
}


void StringLitNode::resolveRequires( ClassDecl *currentClass )
{
  currentClass->requires( *NativeUtilsDecl );
}


////////////////////////////////////////////////////////////////////////


void ClassDecl::requires( TypeNode &prerequisite )
{
  requires( *prerequisite.decl() );
}


void ClassDecl::requires( Decl &prerequisite )
{
  _requirements.insert( &prerequisite );
}


void Decl::requires( TypeNode & )
{
  invalidOperation( "requires" );
}


void Decl::requires( Decl & )
{
  invalidOperation( "requires" );
}
