/* Code for computing and reporting statistics.  Turned on by -stats
   command line flag. */

#include "AST.h"
#include "stats.h"
#include "code-util.h"
#include <cstdio>
#include <iostream.h>
#include <stdio.h>

bool stat_foreach = false;

static int SRorOSR(TreeNode *t, bool local)
{
  if (isDummyNode(t))
    return 0;
  if (isOSRArrayAccessNode(t) ||isSRArrayAccessNode(t))
    return (t->array()->type()->isLocal() == local) ? 1 : 0;
  int count = 0;
  for (int i = t->arity(); i-- > 0; )
    count += SRorOSR(t->child(i), local);
  return count;
}

static int OSR(TreeNode *t, bool local)
{
  if (isDummyNode(t))
    return 0;
  if (isOSRArrayAccessNode(t))
    return (t->array()->type()->isLocal() == local) ? 1 : 0;
  int count = 0;
  for (int i = t->arity(); i-- > 0; )
    count += OSR(t->child(i), local);
  return count;
}

static int plaingrid(TreeNode *t, bool local)
{
  if (isDummyNode(t))
    return 0;
  if (isTitaniumArrayAccessNode(t))
    return (t->array()->type()->isLocal() == local) ? 1 : 0;
  int count = 0;
  for (int i = t->arity(); i-- > 0; )
    count += plaingrid(t->child(i), local);
  return count;
}

static int java(TreeNode *t)
{
  if (isDummyNode(t))
    return 0;
  if (isJavaArrayAccessNode(t))
    return 1;
  int count = 0;
  for (int i = t->arity(); i-- > 0; )
    count += java(t->child(i));
  return count;
}

static TreeNode *child0(TreeNode *x)
{
  return (x->arity() > 0) ? x->child(0) : x;
}

// t is the tree to search for uses of array.  d is the type of the domain
// we're iterating over (or NULL if not a foreach loop).
void loopStats(TreeNode *t, TypeNode *d, ostream &os)
{
  char q[100];
  static bool inited = false;

  if (!inited) {
    os <<
      "How to interpret loop stats:\n"
      " The first line of the loop is given in parenthesis before a colon.\n"
      " The numbers are reported in three groups.  The first three numbers are for\n"
      " local grids.  The next three are for global grids.  The last\n"
      " is for Java arrays.  The three numbers for grids are as follows:\n"
      " `SR(OSR) Regular.'  SR is the number of strength reduced array"
      " accesses.\n OSR is the number of those that are offset strength\n"
      " reduced.  Regular is the number that are not optimized.  Accesses\n"
      " to Java arrays are always local and are never strength reduced so"
      " there is\n just one number.  At the end of the line is 'Rn' or 'Dn'"
      " where is n is\n an arity and 'R' and 'D' indicate RectDomain and"
      " Domain, respectively.\n\n";
    inited = true;
  }

  // Guestimate a node to tell us the source code position of the loop.
  TreeNode *p = (d == NULL) ? t : child0(child0(t->parent()));

  string s = "Loop stats (" + p->position().asString() + "): ";
  sprintf(q, "%s", s.c_str());
  while (strlen(q) < 46)
    sprintf(q + strlen(q), " ");
  sprintf(q + strlen(q), "%2d(%2d) %2d | %2d(%2d) %2d | %2d",
	  SRorOSR(t, true), OSR(t, true), plaingrid(t, true),
	  SRorOSR(t, false), OSR(t, false), plaingrid(t, false),
	  java(t)); 
  os << q;
  if (d != NULL)
    os << (d->isRectDomainType() ? " | R" : " | D")
       << int2string(d->tiArity());  
  os << endl;
}
