// Strange fiddlings required to use a nonstandard basic_string with g++

#include <iostream>
#include <iomanip>
#include "config.h"
#include "string16.h"

#ifdef _AIX

// Why's this here?
// C++ in its infinite complexity lets you perform a generic initialization
// of static members of parametrized classes, e.g.
//
// template <class foo> 
// int someclass<foo>::i = 666;
//
// but old versions of egcs, like the one they have on the SDSC SP/2, require
// initialization of the instantiation instead:
// int someclass<int>::i = 666;

string16::Rep string16::nilRep = { 0, 0, 1, false };

#endif

// This representation is suitable for human viewing,
// but cannot be fed into a C compiler.

ostream &operator<<( ostream &os, const string16 &string )
{
    const ios::fmtflags flags = os.setf( ios::hex, ios::basefield | ios::showbase );
    const char fill = os.fill( '0' );
    os << '"';
    
    for (unsigned pos = 0; pos < string.length(); ++pos) {
	const uint16 character = string[ pos ];
	switch (character) {
	case '\b': os << "\\b"; break;
	case '\t': os << "\\t"; break;
	case '\n': os << "\\n"; break;
	case '\f': os << "\\f"; break;
	case '\r': os << "\\r"; break;
	case '"' : os << "\\\""; break;
	case '\\': os << "\\\\"; break;
	default:
	    if (isascii( character ) && isprint( character ))
		os << (char) character;
	    else
		os << "\\u" << setw( 4 ) << character;
	}
    }

    os << '"';
    os.fill( fill );
    os.flags( flags );
    return os;
}


// This representation may be fed into a C compiler,
// but is not as good for human viewing.

void printEscaped( ostream &os, uint16 character )
{
    const ios::fmtflags flags = os.setf( ios::hex | ios::showbase, ios::basefield | ios::showbase );
    const char fill = os.fill( '0' );

    switch (character) {
    case '\b': os << "'\\b'"; break;
    case '\t': os << "'\\t'"; break;
    case '\n': os << "'\\n'"; break;
    case '\f': os << "'\\f'"; break;
    case '\r': os << "'\\r'"; break;
    case '\'': os << "'\\''"; break;
    case '"' : os << "'\\\"'"; break;
    case '\\': os << "'\\\\'"; break;
    default:
	if (isascii( character ) && isprint(character))
	    os << '\'' << char( character ) << '\'';
	else
	    os << character;
    }

    os.fill( fill );
    os.flags( flags );
}


// This representation may be fed into a C compiler,
// but is not as good for human viewing.

void printEscaped( ostream &os, const string16 &string )
{
    os << "{ ";
    bool first = true;
    
    for (unsigned pos = 0; pos < string.length(); ++pos) {
	if (first) first = false; else os << ", ";
	printEscaped( os, string[ pos ] );
    }

    os << " }";
}
